"""Neural network and model package.

This package contains classes that define blocks and modules used in various neural network
architectures. Most of these classes have been adapted from external sources; see their individual
headers for more information.
"""

import logging

import thelper.nn.utils  # noqa: F401
from thelper.nn.utils import Module  # noqa: F401
from thelper.nn.utils import create_model  # noqa: F401

logger = logging.getLogger("thelper.nn")
